#include <device_launch_parameters.h> // cuda keywords
#include <cuda_runtime.h> // cuda runtime functions

#include <stdio.h>
#include <iostream>
#include <stdlib.h>
#include <random>

// Simple utility function to check for CUDA runtime errors, forward declaration
void checkCUDAError(const char *msg)
{
    cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err)
    {
        fprintf(stderr, "Cuda error: %s: %s.\n", msg, cudaGetErrorString( err ) );
        exit(-1);
    }
}

// ** TODO ** exercise 2.d: implement the kernel body
__global__ void dotProduct( int *d_in1, int *d_in2, int *d_out )
{
    unsigned int tid = blockDim.x * blockIdx.x + threadIdx.x;
    int cache = d_in1[tid] * d_in2[tid];
    atomicAdd( d_out, cache );
}

int main( int argc, char** argv)
{
    // pointer for host memory
    int *h_in1, *h_in2, *h_out;

    // pointer for device memory
    int *d_in1, *d_in2, *d_out;

    // define grid and block size
    const unsigned int numBlocks = 3;
    const unsigned int numThreadsPerBlock = 3;
    const unsigned int numElements = numBlocks * numThreadsPerBlock;

    h_in1 = new int[numElements];
    h_in2 = new int[numElements];
    h_out = new int[1];

    // init vector data
    h_out[0] = 0;
    std::random_device rd;
    std::mt19937 gen(rd());
    std::uniform_real_distribution<> dis(-100.0, 100.0);
    for (unsigned int i=0; i<numElements; i++)
        h_in1[i] = static_cast<int>( std::round(dis(gen)) );
    for (unsigned int i=0; i<numElements; i++)
        h_in2[i] = static_cast<int>( std::round(dis(gen)) );

    // copy host arrays to device arrays
    cudaMalloc( &d_in1, numElements*sizeof(int) );
    cudaMalloc( &d_in2, numElements*sizeof(int) );
    cudaMalloc( &d_out, 1*sizeof(int) );
    checkCUDAError( "cudaMalloc" );

    // copy host arrays to device arrays
    cudaMemcpy( d_in1, h_in1, numElements*sizeof(int), cudaMemcpyHostToDevice );
    cudaMemcpy( d_in2, h_in2, numElements*sizeof(int), cudaMemcpyHostToDevice );
    cudaMemcpy( d_out, h_out, 1*sizeof(int), cudaMemcpyHostToDevice );
    checkCUDAError( "cudaMemcpy" );

    // configure and launch the kernel
    dim3 dimGrid( numBlocks, 1, 1 );
    dim3 dimBlock( numThreadsPerBlock, 1, 1 );
    dotProduct<<<dimGrid, dimBlock>>>( d_in1, d_in2, d_out );

    // block until the device has completed
    cudaDeviceSynchronize();
    checkCUDAError( "Kernel execution" );

    // copy device arrays to host arrays
    cudaMemcpy( h_out, d_out, 1*sizeof(int), cudaMemcpyDeviceToHost );
    checkCUDAError( "cudaMemcpy" );

    std::cout << std::endl << "Input 1: [ ";
    for (unsigned int i=0; i<numElements; i++)
        std::cout << h_in1[i] << " ";
    std::cout << "]" << std::endl;
    std::cout << std::endl << "Input 2: [ ";
    for (unsigned int i=0; i<numElements; i++)
        std::cout << h_in2[i] << " ";
    std::cout << "]" << std::endl;
    std::cout << std::endl << "Result: " << *h_out << std::endl;

    // free host memory
    delete[] h_in1;
    delete[] h_in2;
    delete[] h_out;

    // free device memory
    cudaFree( d_in1 );
    cudaFree( d_in2 );
    cudaFree( d_out );

    return 0;
}
